/*---------------------------------------------
                 S I M P A C K
            Simulation Tool Package
        Copyright 1990, Paul A. Fishwick
---------------------------------------------*/
/* This is a set of routines for discrete event simulation.
It was inspired by both CSIM (Herb Schwetman) and SMPL
(M. H. MacDougall). The syntax is very similar to SMPL's
with the exceptions as noted in the file README.
*/

/*---------------------------------------------
Modified by Gheith Abandah in Sep 1997, list of modifications:

  1. Time is changed from float into unsigned long.
  2. tokens changed from float into unsigned long.
  3. Preemption is deleted.
  4. Simplified to get faster simulation.
  5. Performance vars from float into double.
  6. Only heap and linked are supported.
  7. Remove duplicats is not supported.
---------------------------------------------*/

#include <malloc.h>
#include <stdio.h> 
#include <math.h>
//#include <curses.h>
#define TRUE 1
#define FALSE 0
//

#include <assert.h>

#include "queuing.h"

typedef struct item{
  TIME time;
  int event;
  TOKEN token;
  int priority;
} ITEM;

typedef struct node {
  ITEM item;
  struct node *next;
} NODE;

typedef struct {
  NODE *front;
  int size;
} LIST;

typedef struct {
  LIST queue;
  int status;
  char name[25];
  int total_servers;
  int busy_servers;
  TIME total_busy_time;
  TIME start_busy_time;
  int server_info[MAX_SERVERS][2];
} FACILITY;

TIME last_event_time;
TIME total_token_time;
int current_event, facilities, arrivals, completions;
int tokens_in_system, trace_flag, trace_type, heap_count;
int event_list_type;
char current_operation[100];
LIST event_list;
ITEM heap[HEAP_SIZE], an_item,item1;
FACILITY facility[MAX_FACILITIES];

/* Performance variables */
double utilization, idle, arrival_rate, throughput, total_sim_time;
double total_busy_time;
double total_utilization;
double mean_service_time, mean_num_tokens, mean_residence_time;

/* Declarations for random distribution sampling */

#define then

#ifdef INCLUDE_RANDOM
#define A 16807L           /* multiplier (7**5) for 'ranf' */
#define M 2147483647L      /* modulus (2**31-1) for 'ranf' */

long In[] = {0L,           /* seeds for streams 1 thru 15  */
  1973272912L,  747177549L,   20464843L,  640830765L, 1098742207L,
    78126602L,   84743774L,  831312807L,  124667236L, 1172177002L,
  1124933064L, 1223960546L, 1878892440L, 1449793615L,  553303732L};

int strm=1;                /* index of current stream */
#endif

int endwin () {}
int clear () {}
int move (int x, int y) {}
int refresh (){}
int initscr (){}

/* -------------------------------------------------------------------- */


/************************************************************************/
/*                                                                      */
/* init_simpack:                                                        */
/*      Initialize data structures and variables for simulation.        */
/*                                                                      */
/************************************************************************/
init_simpack (flags)
int flags;
{
  static int rn_stream = 1;
  int i, type;
  int ranmark;

  type = flags & 1;
  ranmark = flags & 32;
  /* initialize token list table */
  event_list_type = type;
  create_list(&event_list);
  heap_count = 0;
  trace_flag = OFF;
  current_time = 0;
  last_event_time = current_time;
  facilities = 0;
  arrivals = 0;
  completions = 0;
  total_token_time = 0;
  tokens_in_system = 0;

#ifdef INCLUDE_RANDOM
  /* set random number stream */
  rn_stream = stream(rn_stream);
/* this next part is included so that one may run multiple passes
   with different data structures for the FEL but generate the same
   random number stream for easy comparison.  To use this, call
   init_simpack with the flags binary or'ed with RESEED  */
  if (ranmark) {
    In[0] = 0L;
    In[1] = 1973272912L;
    In[2] = 747177549L;
    In[3] = 20464843L;
    In[4] = 640830765L;
    In[5] = 1098742207L;
    In[6] = 78126602L;
    In[7] = 84743774L;
    In[8] = 831312807L;
    In[9] = 124667236L;
    In[10] = 1172177002L;
    In[11] = 1124933064L;
    In[12] = 1223960546L;
    In[13] = 1878892440L;
    In[14] = 1449793615L;
    In[15] = 553303732L;
  }
  else rn_stream++;
  if (rn_stream > 15) rn_stream = 1;
#endif
} /* end init() */


end_simpack()
{
  int i;
  ITEM an_item;
  switch (event_list_type) {
  case LINKED: listblankqueue(&(event_list.front)); break;
  case HEAP: break;
  }
  for (i=facilities;i>0;i--) {
    while (!empty_list(&(facility[i].queue)))
      remove_front_list(&(facility[i].queue),&an_item);
  } 
}

/************************************************************************/
/*                                                                      */
/* time:                                                                */
/*      Provide the current simulation time.                            */
/*                                                                      */
/************************************************************************/
TIME sim_time()
{
  return(current_time);
} /* end time() */




/************************************************************************/
/*                                                                      */
/* busy_time:                                                           */
/*      Provide total accumulated busy time for a facility.             */
/*                                                                      */
/************************************************************************/
TIME busy_time(facility_id)
int facility_id;
{
  return(facility[facility_id].total_busy_time); 
}




/************************************************************************/
/*                                                                      */
/* update_arrivals:                                                     */
/*      Update # of arrivals into a system.                             */
/*                                                                      */
/************************************************************************/
update_arrivals()
{
  arrivals++;
  tokens_in_system++;
}




/************************************************************************/
/*                                                                      */
/* update_completions:                                                  */
/*      Update # of completions out of a system.                        */
/*                                                                      */
/************************************************************************/
update_completions()
{
  completions++;
  tokens_in_system--;
}




/************************************************************************/
/*                                                                      */
/* facility_status:                                                     */
/*      Return facility status.                                         */
/*                                                                      */
/************************************************************************/
int facility_status(facility_id)
int facility_id;
{
  return(facility[facility_id].status);
}

/************************************************************************/
/*                                                                      */
/* facility_qsize:                                                      */
/*      Return facility queue size.                                     */
/*                                                                      */
/************************************************************************/
int facility_qsize(facility_id)
int facility_id;
{
  return(facility[facility_id].queue.size);
}




/************************************************************************/
/*                                                                      */
/* create_facility:                                                     */
/*      Setup a facility with a given number of servers.                */
/*                                                                      */
/************************************************************************/
int create_facility(name,num_servers)
char *name;
int num_servers;
{ int i;

  facilities++;
  assert (facilities < MAX_FACILITIES);

  create_list(&facility[facilities].queue);
  facility[facilities].status = FREE;
  strcpy(facility[facilities].name,name);
  facility[facilities].total_servers = num_servers;
  facility[facilities].busy_servers = 0;

  /* clear servers */
  for (i=1;i<=num_servers;i++) {
   facility[facilities].server_info[i][0] = 0;
   facility[facilities].server_info[i][1] = 0;
  } /* end for */

  facility[facilities].total_busy_time = 0;
  return(facilities);
} /* end facility() */




/************************************************************************/
/*                                                                      */
/* schedule:                                                            */
/* Schedule a given event in a given time.                              */
/*                                                                      */
/************************************************************************/
schedule(event,inter_time,token)
int event;
TIME inter_time;
TOKEN token;
{
  TIME event_time;
  ITEM an_item;
  int i, token_id;

  event_time = current_time + inter_time;

#ifdef TRACE_EVENTS
                                                   /* update trace if active */
    sprintf(current_operation,
            "BEFORE SCHEDULE Event %2d Inter-Time %d Token %d",
            event,inter_time, token.attr[0]);
    trace_update();
#endif

    an_item.time = event_time;
    an_item.event = event;
    for (i=0;i<MAX_NUM_ATTR;i++)
      an_item.token.attr[i] = token.attr[i];

    switch (event_list_type) {
    case LINKED: insert_list(&event_list,&an_item,TIME_KEY); break;
    case HEAP: heap_insert(an_item); break;
    }


                                                   /* update trace if active */
#ifdef TRACE_EVENTS
    sprintf(current_operation,
            "AFTER SCHEDULE Event %2d Inter-Time %d Token %d",
            event,inter_time, token.attr[0]);
    trace_update();
#endif
} /* end schedule() */




/************************************************************************/
/*                                                                      */
/* next_event:                                                          */
/* Cause an event to occur and return it in the given pointers.         */
/*                                                                      */
/************************************************************************/
next_event(event_ptr,token_ptr)
int *event_ptr;
TOKEN *token_ptr;
{
  ITEM an_item;


#ifdef TRACE_EVENTS
                                                   /* update trace if active */
  strcpy(current_operation,"BEFORE NEXT_EVENT");
  trace_update();
#endif

  switch (event_list_type) {
  case LINKED: remove_front_list(&event_list,&an_item); break;
  case HEAP: heap_remove(&an_item); break;
  }
  current_time = an_item.time;
  current_event = an_item.event;
  *event_ptr = an_item.event;
  *token_ptr = an_item.token;

                                                   /* update trace if active */
#ifdef TRACE_EVENTS
  sprintf(current_operation,"AFTER NEXT_EVENT Event %2d Token %d",
          *event_ptr, (*token_ptr).attr[0]);
  trace_update();
#endif

                      /* update the total token-time spent within the system */
  total_token_time += tokens_in_system*(sim_time() - last_event_time);
  last_event_time = sim_time();
} /* end next_event() */




/************************************************************************/
/*                                                                      */
/* request:                                                             */
/*      Request the use of a given queue.                               */
/*                                                                      */
/************************************************************************/
int request(facility_id,token,priority)
int facility_id,priority;
TOKEN token;
{
 ITEM an_item;
 int i,server_num;

#ifdef TRACE_EVENTS
                                                   /* update trace if active */
  sprintf(current_operation,
	  "BEFORE REQUEST Facility %2d Token %d Priority %2d",
          facility_id, token.attr[0],priority);
  trace_update();
#endif

  if (facility[facility_id].busy_servers == 0)
      facility[facility_id].start_busy_time = sim_time();

  if (facility[facility_id].status == FREE) {
    facility[facility_id].busy_servers++;

    /* search for an open server */
    server_num = 1;
    while (facility[facility_id].server_info[server_num][0] != 0) 
          server_num++;
    facility[facility_id].server_info[server_num][0] = token.attr[0];
    facility[facility_id].server_info[server_num][1] = priority;
    if (facility[facility_id].busy_servers ==
       facility[facility_id].total_servers)
         facility[facility_id].status = BUSY;
#ifdef TRACE_EVENTS
                                                   /* update trace if active */
    sprintf(current_operation,
	    "AFTER REQUEST Facility %2d Token %d Priority %2d",
            facility_id, token.attr[0],priority);
    trace_update();
#endif
    return(FREE);
  } else {                            /* facility is BUSY (all servers busy) */
    an_item.time = current_time; /* time that token entered fac. queue */
    for (i=0;i<MAX_NUM_ATTR;i++)
      an_item.token.attr[i] = token.attr[i];
    an_item.priority = priority;
    an_item.event = current_event;
    insert_list(&facility[facility_id].queue,&an_item,BEHIND_PRIORITY_KEY);
#ifdef TRACE_EVENTS
                                                   /* update trace if active */
    sprintf(current_operation,
	    "AFTER REQUEST Facility %2d Token %d Priority %2d",
	    facility_id, token.attr[0],priority);
    trace_update();
#endif
    return(BUSY);
  }
} /* end request() */



/************************************************************************/
/*                                                                      */
/* release:                                                             */
/*      Release server to waiting tokens (if any)                       */
/*                                                                      */
/************************************************************************/
release(facility_id,token)
 int facility_id;
 TOKEN token;
{
   ITEM an_item;
   int server_num,i,found;

#ifdef TRACE_EVENTS
   /* update trace if active */
   sprintf(current_operation,"BEFORE RELEASE Facility %2d Token %d",
          facility_id, token.attr[0]);
   trace_update();
#endif
   server_num = facility[facility_id].total_servers;
   found = FALSE;
   i = 1;
   while(!found && i<=server_num) {
      if(facility[facility_id].server_info[i][0] == token.attr[0]) {
	 facility[facility_id].server_info[i][0] = 0;
	 facility[facility_id].server_info[i][1] = 0;
	 found = TRUE;
      } /* end if */
   i++;
   } /* end while */
   if(found) {
      facility[facility_id].status = FREE;
      facility[facility_id].busy_servers--;
      if(facility[facility_id].busy_servers == 0) {
         facility[facility_id].total_busy_time += sim_time() - 
         	    facility[facility_id].start_busy_time;
         facility[facility_id].start_busy_time = sim_time();
      } /* end if */
      if(facility[facility_id].queue.size > 0) {
      /* at least one entity was waiting on this facility/server */
      remove_front_list(&facility[facility_id].queue,&an_item);
        an_item.time = current_time;
	switch (event_list_type) {
	case LINKED: add_front_list(&event_list,&an_item); break;
	case HEAP: heap_insert(an_item); break;
	}
    } /* end if */
  } /* end if (found) */

#ifdef TRACE_EVENTS
                                                   /* update trace if active */
  sprintf(current_operation,"AFTER RELEASE Facility %2d Token %d",
          facility_id, token.attr[0]);
  trace_update();
#endif

} /* end release() */




/************************************************************************/
/*                                                                      */
/* create_list:                                                         */
/*      Create a list structure.                                        */
/*                                                                      */
/************************************************************************/
create_list(list_ptr)
  LIST *list_ptr;
{
  list_ptr->front = NIL;
  list_ptr->size = 0;
} /* end create_list() */




/************************************************************************/
/*                                                                      */
/* empty_list:                                                          */
/*      Answers the question -- "is the list empty?"                    */
/*                                                                      */
/************************************************************************/
empty_list(list_ptr)
  LIST *list_ptr;
{
  return(list_ptr->size == 0);
} /* end empty_list() */




/************************************************************************/
/*                                                                      */
/* insert_list:                                                         */
/*      Insert an item inside the list according to a key.              */
/*                                                                      */
/************************************************************************/
insert_list(list_ptr,item_ptr,key)
  LIST *list_ptr;
  ITEM *item_ptr;
  int key;
{
  NODE *current_node,*previous_node;
  NODE *new_node;
 
  new_node = (NODE *) calloc(1,sizeof(NODE));
  new_node->item = *item_ptr;
  new_node->next = NIL;
  list_ptr->size += 1;
  /* printf("size = %d\n",facility[1].queue.size); */

  current_node = list_ptr->front;
  previous_node = list_ptr->front;
                                   /* insert in list according to sorted key */
  if (key == TIME_KEY) 
    while ((current_node != NIL) && 
           (current_node->item.time <= item_ptr->time)) {
      previous_node = current_node;
      current_node = current_node->next;
    } /* end while */
  else if (key == BEHIND_PRIORITY_KEY)
    while ((current_node != NIL) && 
           (current_node->item.priority >= item_ptr->priority)) {
      previous_node = current_node;
      current_node = current_node->next;
    } /* end while */
    else /* key == AHEAD_PRIORITY_KEY */
      while ((current_node != NIL) && 
             (current_node->item.priority > item_ptr->priority)) {
        previous_node = current_node;
        current_node = current_node->next;
      } /* end while */

  if (current_node == NIL) {
                                     /* add the inserted item to end of list */
    if (previous_node == NIL) /* list is empty */
      list_ptr->front = new_node;
    else
      previous_node->next = new_node;
  } else {
                                     /* insert item between 2 existing items */
    if (previous_node != current_node) {
                                                      /* list has >= 2 items */
      previous_node->next = new_node;
      new_node->next = current_node;
    } 
                                                     /* list has only 1 item */
    else {
      list_ptr->front = new_node;
      new_node->next = current_node;
    }
  }
} /* end insert_list() */




/************************************************************************/
/*                                                                      */
/* add_front_list:                                                      */
/*      Add item at front of list.                                      */
/*                                                                      */
/************************************************************************/
add_front_list(list_ptr,item_ptr)
  LIST *list_ptr;
  ITEM *item_ptr;
{
  NODE *new_node;
 
  new_node = (NODE *) calloc(1,sizeof(NODE));
  new_node->item = *item_ptr;
  new_node->next = NIL;
  if (list_ptr->size == 0) 
    list_ptr->front = new_node;
  else {
    new_node->next = list_ptr->front;
    list_ptr->front = new_node;
  }   
  list_ptr->size += 1;
                            /* printf("size = %d\n",facility[1].queue.size); */
} /* end add_front_list() */




/************************************************************************/
/*                                                                      */
/* remove_front_list:                                                   */
/*      Remove item from front of list.                                 */
/*                                                                      */
/************************************************************************/
remove_front_list(list_ptr,item_ptr)
  LIST *list_ptr;
  ITEM *item_ptr;
{
  NODE *temp_ptr;

  *item_ptr = list_ptr->front->item;
  temp_ptr = (NODE *) list_ptr->front;
  list_ptr->front = list_ptr->front->next;
                                                    /* deallocate node space */
  free(temp_ptr); 
  list_ptr->size -= 1;
} /* end remove_front_list() */


/* added by D. Hay */
listblankqueue(q)
NODE *q;
{
  if (q==NULL) return;
  listblankqueue(q->next);
  free(q);
}

/************************************************************************/
/*                                                                      */
/* trace_facility:                                                      */
/*      Output information about the given facility.                    */
/*                                                                      */
/************************************************************************/
trace_facility(facility_id)
int facility_id;
{
  NODE *node;
  LIST *list_ptr;
  int i;

  list_ptr = &facility[facility_id].queue;
  printf("Time: %d\n",sim_time());
  printf("Queue %d: ",facility_id);
  node = list_ptr->front;
  while (node != NIL) {
    printf("(TM %d,",node->item.time);
    printf("EV %1d,",node->item.event);
    printf("TK %d) ", node->item.token.attr[0]);
    node = node->next;
  }
  printf("\n");
  for(i=0;i<60;i++) printf("-");
  printf("\n");
} /* end trace_facility() */




/************************************************************************/
/*                                                                      */
/* trace_eventlist:                                                     */
/*      Output information about the eventlist.                         */
/*                                                                      */
/************************************************************************/
trace_eventlist()
{
  NODE *node;
  LIST *list_ptr;
  int i;

  list_ptr = &event_list;
  printf("Time: %d\n",sim_time());
  printf("Events: ");
  node = list_ptr->front;
  while (node != NIL) {
    printf("(TM %d,",node->item.time);
    printf("EV %1d,",node->item.event);
    printf("TK %d) ", node->item.token.attr[0]);
    node = node->next;
  }
  printf("\n");
  for(i=0;i<60;i++) printf("-");
  printf("\n");
} /* end trace_eventlist() */








/************************************************************************/
/*                                                                      */
/* trace_eventlist_size:                                                */
/*      Output the size of the event list.                              */
/*                                                                      */
/************************************************************************/
int trace_eventlist_size()
{
  return(event_list.size);
} /* end trace_eventlist_size() */




/************************************************************************/
/*                                                                      */
/* trace_visual:                                                        */
/*      Complete Graphical Trace of Queuing Simulation.                 */
/*                                                                      */
/************************************************************************/
trace_visual(type)
int type;
{
  trace_type = type;
  trace_flag = ON;
  strcpy(current_operation,"START SIMULATION");
  /* initialize screens */
  if (trace_type == INTERACTIVE) initscr();
}




/************************************************************************/
/*                                                                      */
/* trace_update:                                                        */
/*      Output a detailed graphical trace.                              */
/*                                                                      */
/************************************************************************/
trace_update()
{
  LIST *list_ptr;
  NODE *node;
  int i,j;
  char any_char;

  if (trace_flag == ON && trace_type == INTERACTIVE) {
    /* Refresh screen */
    clear();
    refresh();
    move(0,0);
  }

  if (trace_flag == ON) {

  printf("## TIME: %d\n",sim_time()); 
  printf("## %s\n",current_operation);
  printf("\n");

  if(event_list_type == LINKED) {

                                                         /* Trace Event List */
    printf("## EVENT LIST\n");

    list_ptr = &event_list;
    printf("      ");  
    node = list_ptr->front;
    while (node != NIL) {
      printf("  ");
      printf("+-----+");
      node = node->next;
    }
    printf("\n");

    printf("Token ");
    node = list_ptr->front;
    while (node != NIL) {
      printf("  ");
      printf("|%5d|", node->item.token.attr[0]);
      node = node->next;
    }
    printf("\n");

    printf("Time  ");  
    node = list_ptr->front;
    while (node != NIL) {
      printf("<=");
      printf("|%5d|",node->item.time);
      node = node->next;
    }
    printf("\n");

    printf("Event ");  
    node = list_ptr->front;
    while (node != NIL) {
      printf("  ");
      printf("|%5d|",node->item.event);
      node = node->next;
    } 
    printf("\n");

    printf("      ");  
    node = list_ptr->front;
    while (node != NIL) {
      printf("  ");
      printf("+-----+");
      node = node->next;
    }
    printf("\n");
  } else {
                                                       /* Trace HEAP */
	printf("## PRIORITY QUEUE\n");
	print_heap();
  }
                                                /* Trace each facility queue */
  for(i=1;i<=facilities;i++) {

    printf("## FACILITY %2d: (%s). %2d Server(s), %2d Busy.\n",
	   i,facility[i].name,facility[i].total_servers,
	   facility[i].busy_servers);
    printf("Server(s): ");
    for(j=1;j<=facility[i].total_servers;j++)
       printf("(%2d) TK %3d PR %2d ",j,facility[i].server_info[j][0],
	      facility[i].server_info[j][1]);
    printf("\n");

    list_ptr = &facility[i].queue;
    printf("         ");  
    node = list_ptr->front;
    while (node != NIL) {
      printf("  ");
      printf("+-----+");
      node = node->next;
    }
    printf("\n");

    printf("Token    ");
    node = list_ptr->front;
    while (node != NIL) {
      printf("  ");
      printf("|%5d|", node->item.token.attr[0]);
      node = node->next;
    }
    printf("\n");
    
    printf("Time     ");  
    node = list_ptr->front;
    while (node != NIL) {
      printf("<=");
      printf("|%5d|",node->item.time);
      node = node->next;
    }
    printf("\n");

    printf("Event    ");  
    node = list_ptr->front;
    while (node != NIL) {
      printf("  ");
      printf("|%5d|",node->item.event);
      node = node->next;
    }
    printf("\n");

    printf("Priority ");  
    node = list_ptr->front;
    while (node != NIL) {
      printf("  ");
      printf("|%5d|",node->item.priority);
      node = node->next;
    }
    printf("\n");

    printf("         ");  
    node = list_ptr->front;
    while (node != NIL) {
      printf("  ");
      printf("+-----+");
      node = node->next;
    } 
    printf("\n");
  } /* end for */

  printf("\n");
  if (trace_type == INTERACTIVE) {
                       /* if INTERACTIVE mode, hold screen until key pressed */
    printf("-- press any key to continue ('X' to exit) --");
    any_char = getchar();
    if (any_char == 'x' || any_char == 'X') {
      endwin();
      printf("\n");
      exit(0);
    }
  } /* end if */
} /* end if (trace_flag == ON) */
} /* end trace_update() */




#ifdef INCLUDE_RANDOM
typedef double real;

/* These random number generator routines were written by
   M.H. MacDougall, "Simulating Computer Systems", MIT Press, 1987.
*/

/*-------------  UNIFORM [0, 1] RANDOM NUMBER GENERATOR  -------------*/
/*                                                                    */
/* This implementation is for Motorola 680x0 CPUs using C compilers   */
/* with 16-bit short integers and 32-bit long integers.               */
/*                                                                    */
/*--------------------------------------------------------------------*/
real ranf_old()
  {
    short *p,*q,k; long Hi,Lo;
    /* generate product using double precision simulation  (comments  */
    /* refer to In's lower 16 bits as "L", its upper 16 bits as "H")  */
    p=(short *)&In[strm]; Hi= *(p)*A;                   /* 16807*H->Hi */
    *(p)=0; Lo=In[strm]*A;                             /* 16807*L->Lo */
    p=(short *)&Lo; Hi+= *(p);      /* add high-order bits of Lo to Hi */
    q=(short *)&Hi;                       /* low-order bits of Hi->LO */
    *(p)= *(q+1)&0X7FFF;                             /* clear sign bit */
    k= *(q)<<1; if (*(q+1)&0X8000) then k++;       /* Hi bits 31-45->K */
    /* form Z + K [- M] (where Z=Lo): presubtract M to avoid overflow */
    Lo-=M; Lo+=k; if (Lo<0) then Lo+=M;
    In[strm]=Lo;
    return((real)Lo*4.656612875E-10);             /* Lo x 1/(2**31-1) */
  }

/*
 * New version of the RNG from simpack-2.22, more portable
 * than the original.
 *
 * hrp@cray.com, 94/05/26
 */

/*-------------  UNIFORM [0, 1] RANDOM NUMBER GENERATOR  -------------*/
/*                                                                    */
/* This implementation should be portable to machines with	      */
/* sizeof(int) >=4.					              */
/*                                                                    */
/*--------------------------------------------------------------------*/
real ranf()
  {
    unsigned int k; long Hi,Lo;
    /* generate product using double precision simulation  (comments  */
    /* refer to In's lower 16 bits as "L", its upper 16 bits as "H")  */
    Hi = ((In[strm] >> 16) & 0xffff) * A; /* 16807*H->Hi */
    Lo = (In[strm] & 0xffff) * A;	/* 16807*L->Lo */
    Hi += Lo >> 16;		/* add high-order bits of Lo to Hi */
    Lo &= 0xffff; Lo += (Hi & 0x7fff) << 16; /* low-order bits of Hi->Lo */
    k = Hi >> 15;
    /* form Z + K [- M] (where Z=Lo): presubtract M to avoid overflow */
    Lo-=M; Lo+=k; if (Lo<0) then Lo+=M;
    In[strm]=Lo;
    return((real)Lo*4.656612875E-10); /* Lo / (2**31-1) */
  }

/*--------------------  SELECT GENERATOR STREAM  ---------------------*/
stream(n)
  int n;
    { /* set stream for 1<=n<=15, return stream for n=0 */
      /* if ((n<0)||(n>15)) then error(0,"stream Argument Error"); */
      if (n) then strm=n;
      return(strm);
    }

/*--------------------------  SET/GET SEED  --------------------------*/
long seed(Ik,n)
  long Ik; int n;
    { /* set seed of stream n for Ik>0, return current seed for Ik=0  */
      /* if ((n<1)||(n>15)) then error(0,"seed Argument Error"); */
      if (Ik>0L) then  In[n]=Ik;
      return(In[n]);
    }

/*------------  UNIFORM [a, b] RANDOM VARIATE GENERATOR  -------------*/
real uniform(a,b)
  real a,b;
    { /* 'uniform' returns a psuedo-random variate from a uniform     */
      /* distribution with lower bound a and upper bound b.           */
      /* if (a>b) then error(0,"uniform Argument Error: a > b"); */
      return(a+(b-a)*ranf());
    }

/*--------------------  RANDOM INTEGER GENERATOR  --------------------*/
random_gen(i,n)
  int i,n;
    { /* 'random' returns an integer equiprobably selected from the   */
      /* set of integers i, i+1, i+2, . . , n.                        */
      /* if (i>n) then error(0,"random Argument Error: i > n"); */
      n-=i; n=(n+1.0)*ranf();
      return(i+n);
    }

/*--------------  EXPONENTIAL RANDOM VARIATE GENERATOR  --------------*/
real expntl(x)
  real x;
    { /* 'expntl' returns a psuedo-random variate from a negative     */
      /* exponential distribution with mean x.                        */
      return(-x*log(ranf()));
    }

/*----------------  ERLANG RANDOM VARIATE GENERATOR  -----------------*/
real erlang(x,s)
  real x,s;
    { /* 'erlang' returns a psuedo-random variate from an erlang      */
      /* distribution with mean x and standard deviation s.           */
      int i,k; real z;
      /* if (s>x) then error(0,"erlang Argument Error: s > x"); */
      z=x/s; k=(int)z*z;
      z=1.0; for (i=0; i<k; i++) z *= ranf();
      return(-(x/k)*log(z));
    }

/*-----------  HYPEREXPONENTIAL RANDOM VARIATE GENERATION  -----------*/
real hyperx(x,s)
  real x,s;
    { /* 'hyperx' returns a psuedo-random variate from Morse's two-   */
      /* stage hyperexponential distribution with mean x and standard */
      /* deviation s, s>x.  */
      real cv,z,p;
      /* if (s<=x) then error(0,"hyperx Argument Error: s not > x"); */
      cv=s/x; z=cv*cv; p=0.5*(1.0-sqrt((z-1.0)/(z+1.0)));
      z=(ranf()>p)? (x/(1.0-p)):(x/p);
      return(-0.5*z*log(ranf()));
    }

/*-----------  TRIANGULAR RANDOM VARIATE GENERATION  -----------*/
real triang(a,c,b)
real a,c,b;
{
 /* triangular distribution with left and right being [a,b] and the
    mode being at point c */
  real sample,point;
  point = (c-a)/(b-a);
  sample = uniform(0.0,1.0);
  if (sample <= point)
    return(sqrt(sample*(b-a)*(c-a)) + a);
  else 
    return(b - sqrt((1.0-sample)*(b-a)*(b-c)));
}

/*-----------------  NORMAL RANDOM VARIATE GENERATOR  ----------------*/
real normal(x,s)
  real x,s;
    { /* 'normal' returns a psuedo-random variate from a normal dis-  */
      /* tribution with mean x and standard deviation s.              */
      real v1,v2,w,z1; static real z2=0.0;
      if (z2!=0.0)
        then {z1=z2; z2=0.0;}  /* use value from previous call */
        else
          {
            do
              {v1=2.0*ranf()-1.0; v2=2.0*ranf()-1.0; w=v1*v1+v2*v2;}
            while (w>=1.0);
            w=sqrt((-2.0*log(w))/w); z1=v1*w; z2=v2*w;
          }
      return(x+z1*s);
  }
#endif



/************************************************************************/
/*                                                                      */
/* report_stats:                                                        */
/*      Output all relavent simulation statistics.                      */
/*                                                                      */
/************************************************************************/
report_stats()
{
  int i;

  printf("\n");
  printf("+---------------------------+\n");
  printf("| SimPack SIMULATION REPORT |\n");
  printf("+---------------------------+\n");
  printf("\n");

  if (completions == 0) completions = 1;
  total_sim_time = sim_time();
  total_busy_time = 0.0;
  for(i=1;i<=facilities;i++) {
        /* update busy time if simulation ends when there is a busy facility */
    if(facility[i].busy_servers > 0) 
      facility[i].total_busy_time += sim_time() - 
      facility[i].start_busy_time;
    total_busy_time += busy_time(i);
  }
  total_busy_time /= facilities;
  total_utilization = total_busy_time/total_sim_time;
  arrival_rate = arrivals/total_sim_time;
  throughput = completions/total_sim_time;
  mean_service_time = total_busy_time/completions;
  mean_num_tokens = total_token_time/total_sim_time;
  mean_residence_time = mean_num_tokens/throughput;

  printf("Total Simulation Time: %f\n",total_sim_time);
  printf("Total System Arrivals: %d\n",arrivals);
  printf("Total System Completions: %d\n",completions);
  printf("\n");
  printf("System Wide Statistics\n");
  printf("----------------------\n");
  printf("System Utilization: %4.1f%%\n",100.0*total_utilization);
  printf("Arrival Rate: %f, Throughput: %f\n",arrival_rate,throughput);
  printf("Mean Service Time per Token: %f\n",mean_service_time);
  printf("Mean # of Tokens in System: %f\n",mean_num_tokens);
  printf("Mean Residence Time for each Token: %f\n",mean_residence_time);
  printf("\n");

  printf("Facility Statistics\n");
  printf("-------------------\n");
  for (i=1;i<=facilities;i++) {
    utilization = (100.0 * busy_time(i))/sim_time();
    idle = 100.0 * (sim_time() - busy_time(i))/sim_time();
    printf("F %d (%s): Idle: %4.1f%%, Util: %4.1f%%, Preemptions: %d\n",
	   i,facility[i].name,idle,utilization, 0); 
  }
}




/************************************************************************/
/*                                                                      */
/* heap_insert:                                                         */
/*      Put the given item in the heap structure.                       */
/*                                                                      */
/************************************************************************/
heap_insert(item)
ITEM item;
{
  int parent,child;

  heap_count++;
  heap[heap_count] = item;
  if(heap_count > 1) {
    child = heap_count;
    parent = child/2;
    while ((heap[parent].time > heap[child].time) && (child > 1)) {
      heap_swap(&heap[parent],&heap[child]);
      child = parent;
      if (child > 1)
        parent = child/2;
    } /* end while */
  } /* end if */
} /* end heap_insert */




/************************************************************************/
/*                                                                      */
/* heap_remove:                                                         */
/*      Return the next item at the top of the heap.                    */
/*                                                                      */
/************************************************************************/
heap_remove(item_addr)
ITEM *item_addr;
{
  int parent,child;
  *item_addr = heap[1];
  heap_swap(&heap[1],&heap[heap_count]);
  heap_count--;
  heap_down_adjust(1);

} /* heap_remove */


/************************************************************************/
/*                                                                      */
/* heap_up_adjust:                                                      */
/*      Adjust the heap going up from the given location.               */
/*                                                                      */
/************************************************************************/
heap_up_adjust(child)
 int child;
{
int parent = child/2; 
 
while (parent>0) {
    if (heap[parent].time < heap[child].time) break; 
    heap_swap(&heap[parent],&heap[child]);
    child = parent; parent /= 2; 
  } 
}
/************************************************************************/
/*                                                                      */
/* heap_down_adjust:                                                    */
/*      Adjust the heap going down from the given location.             */
/*                                                                      */
/************************************************************************/
heap_down_adjust(parent)
 int parent;
{
int child;
  /* this loop copied from old heap_delete routine */
  while (1) {
    if (2*parent > heap_count)
      goto exit;
    else
      child = 2*parent;
    if (child+1 <= heap_count) 
      if (heap[child+1].time < heap[child].time)
        child++;
    if (heap[parent].time < heap[child].time) goto exit;
    heap_swap(&heap[parent],&heap[child]);
    parent = child;
  } /* end while */  
 exit: ;
} 

/************************************************************************/
/*                                                                      */
/* heap_delete:                                                         */
/*      Remove a specific item from the heap.                           */
/*                                                                      */
/************************************************************************/
heap_delete(token_id,item)
int token_id;
ITEM *item;
{
  int i,j,parent,child;

  /* find the item in the heap first */
  i = 1;
  while (heap[i].token.attr[0] != token_id) i++;
  /* save all info. in the heap item to be deleted */
  heap_swap(&heap[i],&heap[heap_count]);
  *item = heap[heap_count];
  heap_count--;

  /* reorganize heap */
  heap_up_adjust(i);
  heap_down_adjust(i);

} /* heap_delete */




/************************************************************************/
/*                                                                      */
/* heap_swap:                                                           */
/*      Switch the positions of two items in the heap.                  */
/*                                                                      */
/************************************************************************/
heap_swap(item1,item2)
ITEM *item1,*item2;
{
  ITEM temp;

  temp = *item1;
  *item1 = *item2;
  *item2 = temp;
}




/************************************************************************/
/*                                                                      */
/* print_heap:                                                          */
/*      Output the contents of the heap.                                */
/*                                                                      */
/************************************************************************/
print_heap()
{
          int i;
          for(i=1;i<=heap_count;i++)
	         printf("%d ",heap[i].time);
          printf("\n");
}




